/* r2test --- test program for r2.h, r2x2.h  */
/* Last edited on 2007-01-14 02:28:49 by stolfi */

#include <r2.h>
#include <r2x2.h>

#include <affirm.h>
#include <jsrandom.h>
#include <flt.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define N 2

void check_eq(double x, double y, char *msg);
void check_i_eq(int i, double x, double y, char *msg);
void check_ij_eq(int i, int j, double x, double y, char *msg);
void check_eps(double x, double y, double eps, char *msg);

void check_eq(double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** %20.16e %20.16e\n", (x), (y));
        affirm(0, msg);
      }
  }

void check_i_eq(int i, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d] %20.16e %20.16e\n", (i), (x), (y));
        affirm(0, msg);
      }
  }

void check_ij_eq(int i, int j, double x, double y, char *msg)
  { if (x != y)
      { fprintf(stderr, " ** [%d,%d] %20.16e %20.16e\n", (i), (j), (x), (y));
        affirm(0, msg);
      }
  }

void check_eps(double x, double y, double eps, char *msg)
  { if (fabs(x - y) > eps)
      { fprintf(stderr, " ** %20.16e %20.16e %20.16e\n", (x), (y), (eps));
        affirm(0, msg);
      }
  }

/* Internal prototypes */

int main (int argc, char **argv);
void test_r2(int verbose);
void test_r2x2(int verbose);
void throw_matrix(r2x2_t *m);
double frac (double x);

int main (int argc, char **argv)
  {
    int i;
    srand(1993);
    srandom(1993);

    for (i = 0; i < 100; i++) test_r2(i < 3);
    for (i = 0; i < 100; i++) test_r2x2(i < 3);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }

double frac (double x)
  { int i = (int)x;
    double f = x - (double)i;
    return (f);
  }

void test_r2(int verbose)
  {
    r2_t a, b, c, d, e, para, perp;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    if (verbose)
      { fprintf(stderr,
          "sizeof(r2_t) = %d  %d*sizeof(double) = %d\n",
          sizeof(r2_t), N, N*sizeof(double)
        );
      }

    /* r2_zero: */
    r2_zero(&a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 0.0, "r2_zero error"); }

    /* r2_all: */
    r2_all(3.14, &a);
    for (i = 0; i < N; i++)
      { check_eq(a.c[i], 3.14, "r2_all error"); }

    /* r2_axis: */
    for (k = 0; k < N; k++)
      { r2_axis(k, &a);
        for (i = 0; i < N; i++)
          { check_eq(a.c[i], (i == k ? 1.0 : 0.0), "r2_axis error"); }
      }

    /* r2_throw_cube */
    r2_throw_cube(&a);
    for (i = 0; i < N; i++)
      { affirm(a.c[i] != a.c[(i+1)%N], "r2_throw error(1)"); 
        affirm(frac(a.c[i]*256.0) != 0.0, "r2_throw error(1)"); 
        affirm((a.c[i] > -1.0) && (a.c[i] < 1.0), "r2_throw error(2)"); 
      }

    /* r2_add: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r2_add(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] + b.c[i], "r2_add error"); }

    /* r2_sub: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r2_sub(&a, &b, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], a.c[i] - b.c[i], "r2_sub error"); }

    /* r2_neg: */
    r2_throw_cube(&a);
    r2_neg(&a, &d);
    for (i = 0; i < N; i++)
      { check_eq(d.c[i], - a.c[i], "r2_neg error"); }

    /* r2_scale: */
    s = drandom();
    r2_throw_cube(&a);
    r2_scale(s, &a, &d);
    for (i = 0; i < N; i++)
      { double zi = s*a.c[i];
        check_eq(d.c[i], zi, "r2_scale error(1)");
      }

    /* r2_mix: */
    s = drandom();
    t = drandom();
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r2_mix(s, &a, t, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = s * a.c[i] + t * b.c[i];
        check_eq(d.c[i], ddi, "r2_mix error");
      }

    /* r2_mix_in: */
    s = drandom();
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    d = b;
    r2_mix_in(s, &a, &d);
    for (i = 0; i < N; i++)
      { double ddi = b.c[i] + s * a.c[i];
        check_eq(d.c[i], ddi, "r2_mix_in error");
      }

    /* r2_weigh: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r2_weigh(&a, &b, &d);
    for (i = 0; i < N; i++)
      { double ddi = a.c[i] * b.c[i];
        check_eq(d.c[i], ddi, "r2_weigh error");
      }

    /* r2_norm, r2_norm_sqr, r2_L1_inf_norm: */
    r2_throw_cube(&a);
    r = r2_norm(&a);
    s = r2_norm_sqr(&a);
    t = r2_L_inf_norm(&a);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < N; i++)
      { double ai = fabs(a.c[i]);
        ss += ai*ai; 
        if (ai > tt) { tt = ai; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r2_norm error");
    check_eps(s, ss, 0.000000001 * ss, "r2_norm_sqr error");
    check_eq(t, tt, "r2_L_inf_norm error");

    /* r2_dist, r2_dist_sqr, r2_L_inf_dist: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r = r2_dist(&a, &b);
    s = r2_dist_sqr(&a, &b);
    t = r2_L_inf_dist(&a, &b);
    ss = 0.0;
    tt = 0.0;
    for (i = 0; i < N; i++)
      { double di = fabs(a.c[i] - b.c[i]);
        ss += di*di; 
        if (di > tt) { tt = di; }
      }
    rr = sqrt(ss);
    check_eps(r, rr, 0.000000001 * rr, "r2_dist error");
    check_eps(s, ss, 0.000000001 * ss, "r2_dist_sqr error");
    check_eq(t, tt, "r2_L_inf_dist error");

    /* r2_dir, r2_L_inf_dir: */
    r2_throw_cube(&a);
    r = r2_dir(&a, &b);
    s = r2_L_inf_dir(&a, &d);
    ss = r2_norm(&a);
    tt = r2_L_inf_norm(&a);
    for (i = 0; i < N; i++)
      { check_eps(b.c[i], a.c[i]/ss, 0.000000001 * ss, "r2_dir error");
        check_eps(d.c[i], a.c[i]/tt, 0.000000001 * tt, "r2_L_inf_dir error");
      }

    /* r2_dot, r2_cos, r2_sin: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r = r2_dot(&a, &b);
    s = r2_sin(&a, &b);
    t = r2_cos(&a, &b);
    mag = sqrt(r2_dot(&a,&a)*r2_dot(&b,&b));
    rr = 0.0;
    for (i = 0; i < N; i++) { rr += a.c[i]*b.c[i]; }
    tt = rr/(r2_norm(&a)*r2_norm(&b));
    check_eps(r, rr, 0.000000001 * mag, "r2_dot error(1)");
    check_eps(t, tt, 0.000000001, "r2_cos error(1)");
    d = a;
    r2_mix_in(-rr/r2_norm_sqr(&b), &b, &d);
    ss = r2_norm(&d)/r2_norm(&a);
    check_eps(s, ss, 0.000000001, "r2_sin error(1)");
    for (i = 0; i < N; i++)
      { r2_axis(i, &a);
        for (j = 0; j < N; j++)
          { r2_axis(j, &b);
            r = r2_dot(&a, &b);
            s = r2_sin(&a, &b);
            t = r2_cos(&a, &b);
            rr = (i == j ? 1.0 : 0.0);
            check_eq(r, rr, "r2_dot error(2)");
            check_eq(t, rr, "r2_dot error(2)");
            check_eq(s, 1.0 - rr, "r2_dot error(2)");
          }
      }

    /* r2_cross: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        double sign = ((i % 2) == 0 ? 1.0 : -1.0);
        int p;
        r2_axis(i0, &a);
        r2_cross(&a, &d);
        r2_axis(i1, &e);
        for (p = 0; p < N; p++)
          { double ep = sign*e.c[p];
            check_eq(d.c[p], ep, "r2_cross error(x)");
          }
      }
    /* Test on random vectors: */
    r2_throw_cube(&a);
    r2_cross(&a, &d);
    mag = r2_norm(&a);
    r = r2_dot(&a, &d);
    check_eps(r, 0.0, 0.00000001 * mag, "r2_cross error(1)");

    /* r2_det: */
    /* Test on basis vectors: */
    for (i = 0; i < N; i++)
      { int i0 = (i + 0) % N;
        int i1 = (i + 1) % N;
        double sign = ((i % 2) == 0 ? 1.0 : -1.0);
        r2_axis(i0, &a);
        r2_axis(i1, &b);
        r = r2_det(&a, &b);
        check_eq(r, sign, "r2_det error(2)");
      }
    /* Test on random vectors: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r = r2_det(&a, &b);
    r2_cross(&a, &e);
    rr = r2_dot(&e, &b);
    mag = r2_norm(&a)*r2_norm(&b);
    check_eps(r, rr, 0.00000001 * mag, "r2_det error(1)");

    /* r2_decomp: */
    r2_throw_cube(&a);
    r2_throw_cube(&b);
    r = r2_decomp(&a, &b, &para, &perp);
    rr = r2_dot(&a, &b)/r2_norm_sqr(&b);  
    check_eps(r, rr, 0.000000001 * (fabs(r) + fabs(rr)), "r2_decomp error(1)");
    r2_add(&para, &perp, &c);
    s = r2_dist(&a, &c);
    affirm (s <= 0.000000001 * r2_norm(&a), "r2_decomp error(2)");
    s = r2_dot(&perp, &b);
    check_eps(s, 0.0, 0.000000001 * r2_norm(&b), "r2_decomp error(3)");
    t = r2_dot(&para, &perp);
    check_eps(t, 0.0, 0.000000001 * r2_norm(&a), "r2_decomp error(4)");

    /* r2_print: */
    if (verbose)
      { r2_throw_cube (&a);
        fprintf(stderr, "a = ");
        r2_print(stderr, &a);
        fputc('\n', stderr);
      }
  }

void test_r2x2(int verbose)
  {
    r2x2_t A, B, C;
    r2_t a, b, c, bb, cc;
    double r, s, t;
    double rr, ss, tt;
    double mag;
    int i, j, k;

    /* Size and allocation: */
    if (verbose)
      { fprintf(stderr,
          "sizeof(r2x2_t) = %d  %d*%d*sizeof(double) = %d\n",
          sizeof(r2x2_t), N, N, N*N*sizeof(double)
        );
        fprintf(stderr, "&B = %08x\n", (unsigned)&B);
        fprintf(stderr, "&A-&B = %d\n", ((unsigned)(&A))-((unsigned)(&B)));
        fprintf(stderr, "&B-&C = %d\n", ((unsigned)(&B))-((unsigned)(&C)));
        fprintf(stderr, "&(B.c) = %08x\n", (unsigned)&(B.c));
        fprintf(stderr, "B.c = %08x\n", (unsigned)(B.c));
        fprintf(stderr, "&(B.c[0]) = %08x\n", (unsigned)&(B.c[0]));
        fprintf(stderr, "B.c[0] = %08x\n", (unsigned)(B.c[0]));
        fprintf(stderr, "&(B.c[0][0]) = %08x\n", (unsigned)&(B.c[0][0]));
      }

    /* Indexing and addressing: */
    for (i = 0; i < N; i++)
      for (j = 0; j < N; j++)
        { double *Aij = &(A.c[i][j]); 
          affirm(Aij = ((double *)&A)+(N*i)+j, "r2x2_t indexing error");
        }

    /* r2x2_zero, r2x2_ident: */
    r2x2_zero(&A);
    r2x2_ident(&B);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { check_eq(A.c[i][j], 0.0, "r2x2_zero error"); 
            check_eq(B.c[i][j], (i == j ? 1.0 : 0.0), "r2x2_ident error");
          }
      }

    /* r2x2_map_row, r2x2_map_col: */
    throw_matrix(&A);
    r2_throw_cube(&a);
    r2x2_map_row(&a, &A, &b);
    r2x2_map_col(&A, &a, &c);
    r2_zero(&bb);
    r2_zero(&cc);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { bb.c[j] += a.c[i] * A.c[i][j];
            cc.c[i] += A.c[i][j] * a.c[j];
          }
      }
    r = r2_dist(&b, &bb);
    affirm(r < 0.000000001 * r2_norm(&bb), "r2_map_row error");
    s = r2_dist(&c, &cc);
    affirm(s < 0.000000001 * r2_norm(&cc), "r2_map_col error");

    /* r2x2_mul: */
    throw_matrix(&A);
    throw_matrix(&B);
    r2x2_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double sum = 0.0;
            for (k = 0; k < N; k++) { sum += A.c[i][k]*B.c[k][j]; }
            check_eps(C.c[i][j], sum, 0.000000001 * fabs(sum),
              "r2x2_mul error"
            );
          }
      }

    /* r2x2_det: */
    throw_matrix(&A);
    for (i = 0; i < N; i++)
      { int k = (i + 1) % N;
        for (j = 0; j < N; j++)
          { /* Check for linearity */
            r = drandom();
            A.c[i][j] = r;
            rr = r2x2_det(&A);

            s = drandom();
            A.c[i][j] = s;
            ss = r2x2_det(&A);

            t = drandom();
            A.c[i][j] = r*(1-t) + s*t;
            tt = r2x2_det(&A);
            mag = fabs(rr) + fabs(ss) + fabs(tt);
            check_eps(tt, rr*(1.0 - t) + ss*t, 000000001 * mag,
              "r2x2_det error(1)"
            );
          }

        /* Row swap test: */
        r = r2x2_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[i][j]; A.c[i][j] = A.c[k][j]; A.c[k][j] = t; }
        rr = r2x2_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r2x2_det error(2)");

        /* Col swap test: */
        r = r2x2_det(&A);
        for (j = 0; j < N; j++)
          { double t = A.c[j][i]; A.c[j][i] = A.c[j][k]; A.c[j][k] = t; }
        rr = r2x2_det(&A);
        mag = fabs(r) + fabs(rr);
        check_eps(r, -rr, 000000001 * mag, "r2x2_det error(3)");
      }

    /* r2x2_inv: */
    throw_matrix(&A);
    r2x2_inv(&A, &B);
    r2x2_mul(&A, &B, &C);
    for (i = 0; i < N; i++)
      { for (j = 0; j < N; j++)
          { double val = (i == j ? 1.0 : 0.0);
            affirm((C.c[i][j] - val) < 000000001, "r2x2_inv error");
          }
      }

    /* r2x2_print: */
    if (verbose)
      { throw_matrix (&A);
        fprintf(stderr, "A = ");
        r2x2_print(stderr, &A);
        fputc('\n', stderr);
      }
  }  

void throw_matrix(r2x2_t *m)
  {
    int i, j;
    r2_t a;
    for (i = 0; i < N; i++)
      { r2_throw_cube(&a);
        for (j = 0; j < N; j++) { m->c[i][j] = a.c[j]; }
      }
  }
