/* See indexing.h */
/* Last edited on 2007-01-02 18:58:44 by stolfi */
/* Copyright  2003 by Jorge Stolfi, from University of Campinas, Brazil. */
/* See the rights and conditions notice at the end of this file. */

#include <indexing.h>
#include <affirm.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <assert.h>

/* IMPLEMENTATIONS */

ix_pos_t ix_position ( ix_dim_t d, ix_index_t ix[], ix_pos_t bp, ix_step_t st[] )
  { ix_pos_t p = bp;
    /* Should check range {0..ix_MAX_POS} */
    int i; 
    for (i = 0; i < d; i++)
      { p += (ix_pos_t)(ix[i]*st[i]);
        /* Should check range {0..ix_MAX_POS} */
      }
    return p;
  }

ix_pos_t ix_position_C ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[] )
  { ix_pos_t p = 0;
    if (d <= 0) { return p; }
    /* Should check {0 <= ix[d-1] < sz[d-1]} */
    p += ix[d-1];
    /* Should check range {0..ix_MAX_POS} */
    if (d <= 1) { return p; }
    /* Should worry about exceeding {ix_MAX_ABS_STEP}: */
    ix_step_t st = (ix_step_t)sz[d-1];
    int i; 
    for (i = d-2; i > 0; i--) 
      { /* Should check {0 <= ix[i] < sz[i]} */
        p += ix[i]*st;
        /* Should check range {0..ix_MAX_POS} */
        /* Should worry about exceeding {ix_MAX_ABS_STEP} */
        st *= (ix_step_t)sz[i];
      }
    /* Should check {0 <= ix[0] < sz[0]} */
    p += ix[0]*st;
    /* Should check range {0..ix_MAX_POS} */
    return p;
  }
  
ix_pos_t ix_position_F ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[] )
  { ix_pos_t p = 0;
    if (d <= 0) { return p; }
    /* Should check {0 <= ix[0] < sz[0]} */
    p += ix[0];
    /* Should check range {0..ix_MAX_POS} */
    if (d <= 1) { return p; }
    /* Should worry about exceeding {ix_MAX_ABS_STEP}: */
    ix_step_t st = (ix_step_t)sz[0];
    int i; 
    for (i = 1; i < d-1; i++) 
      { /* Should check {0 <= ix[i] < sz[i]} */
        p += ix[i]*st; 
        /* Should check range {0..ix_MAX_POS} */
        /* Should worry about exceeding {ix_MAX_ABS_STEP} */
        st *= (ix_step_t)sz[i];
      }
    /* Should check {0 <= ix[d-1] < sz[d-1]} */
    /* Should check range {0..ix_MAX_POS} */
    p += ix[d-1]*st;
    return p;
  }

/* INVERSE INDEXING */

void ix_indices_C ( ix_dim_t d, ix_pos_t p, ix_size_t sz[], ix_index_t ix[] )
  { /* Should check range {0..ix_MAX_POS} */
    assert(p >= 0);
    int i; 
    for (i = d-1; i >= 0; i--) 
      { /* Should check {sz[i]>0} */
        ix_size_t w = sz[i]; 
        ix[i] = p % w; 
        /* Should check {ix[i] < ix_MAX_INDEX} */
        p /= w;
      }
    assert(p == 0);
  }
  
void ix_indices_F ( ix_dim_t d, ix_pos_t p, ix_size_t sz[], ix_index_t ix[] )
  { /* Should check range {0..ix_MAX_POS} */
    assert(p >= 0);
    int i; 
    for (i = 0; i < d; i++) 
      { /* Should check {sz[i]>0} */
        ix_size_t w = sz[i];
        ix[i] = p % w;
        /* Should check {ix[i] <= ix_MAX_INDEX} */
        p /= w;
      }
    assert(p == 0);
  }

/* INDEX TUPLE MANIPULATION */

void ix_fill ( ix_dim_t d, ix_index_t ix[], ix_index_t val )
  { 
    int i; 
    for (i = 0; i < d; i++) { ix[i] = val; }
  }

void ix_assign ( ix_dim_t d, ix_index_t ix[], ix_index_t val[] )
  { 
    int i; 
    for (i = 0; i < d; i++) 
      { /* Should check {val[i]>0} */ ix[i] = val[i]; }
  }

bool_t ix_assign_min ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[] )
  { 
    int i; 
    bool_t valid = TRUE;
    for (i = 0; i < d; i++) 
      { ix[i] = 0; if (sz[i] == 0) { valid = FALSE; } }
    return valid;
  }
  
bool_t ix_assign_max ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[] )
  { 
    int i; 
    bool_t valid = TRUE;
    for (i = 0; i < d; i++) 
      { ix[i] = sz[i]-1; if (sz[i] == 0) { valid = FALSE; } }
    return valid;
  }

void ix_shift ( ix_dim_t d, ix_index_t ix[], ix_index_t inc[] )
  { 
    int i; 
    for (i = 0; i < d; i++) { /* Should check {inc[i]>-ix[i]} */  ix[i] += inc[i]; }
  }

bool_t ix_is_valid ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[] )
  { 
    int i; 
    for (i = 0; i < d; i++) 
      { if ((ix[i] < 0) || (ix[i] >= sz[i])) { return FALSE; } }
    return TRUE;
  }

/* ELEMENT COUNTING */

bool_t ix_is_empty ( ix_dim_t d, ix_size_t sz[] )
  { int i; 
    for (i = 0; i < d; i++) 
      { if (sz[i] <= 0) { return TRUE; } }
    return FALSE;
  }

ix_elem_count_t ix_num_tuples ( ix_dim_t d, ix_size_t sz[] )
  { uint64_t n = 1;
    int i;
    for (i = 0; i < d; i++) 
      { ix_size_t szi = sz[i]; 
        if (szi < 1) { return 0; }
        if (szi > 1)
          { demand(szi <= ix_MAX_POSITIONS / n, "too many tuples to count");
            n *= szi;
          }
      }
    return n;
  }
  
ix_elem_count_t ix_num_positions ( ix_dim_t d, ix_size_t sz[], ix_step_t st[] )
  { uint64_t n = 1;
    int i;
    for (i = 0; i < d; i++) 
      { ix_size_t szi = sz[i]; 
        if (szi < 1) { return 0; }
        if ((szi > 1) && (st[i] != 0))
          { demand (szi <= ix_MAX_POSITIONS / n, "overflow");
            n *= szi;
          }
      }
    return n;
  }

ix_pos_t ix_min_pos ( ix_dim_t d, ix_size_t sz[], ix_pos_t bp, ix_step_t st[] )
  { ix_pos_t p = bp;
    int i;
    for (i = 0; i < d; i++) 
      { ix_step_t sti = st[i];
        if (sti < 0)
          { ix_size_t szi = sz[i]; 
            assert(szi > 0);
            p += (szi-1)*sti;
          }
      }
    return p;
  }
  
ix_pos_t ix_max_pos ( ix_dim_t d, ix_size_t sz[], ix_pos_t bp, ix_step_t st[] )
  { ix_pos_t p = bp;
    int i;
    for (i = 0; i < d; i++) 
      { ix_step_t sti = st[i];
        if (sti > 0)
          { ix_size_t szi = sz[i]; 
            assert(szi > 0);
            p += (szi-1)*sti;
          }
      }
    return p;
  }


bool_t ix_same_size ( ix_dim_t d, ix_size_t sza[],  ix_size_t szb[] )
  { ix_axis_t i;
    for (i = 0; i < d; i++) 
      { if (sza[i] != szb[i]) { return FALSE; } }
      return TRUE;
  }

bool_t ix_contained ( ix_dim_t d, ix_size_t sza[],  ix_size_t szb[] )
  { ix_axis_t i;
    for (i = 0; i < d; i++) 
      { if (sza[i] > szb[i]) { return FALSE; } }
      return TRUE;
  }

/* STEP MANIPULATION */

void ix_crop 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_size_t skip, 
    ix_size_t keep
  )
  { /* Check axis: */
    demand(i < d, "bad axis"); 
    /* Check and store the new size: */
    demand(skip + keep <= sz[i], "bad range");
    sz[i] = keep;
    if (keep == 0)
      { /* Array becomes empty: */
        (*bp) = 0;
        ix_axis_t k;
        for (k = 0; k < d; k++) { st[k] = 0; }
      }
    else
      { /* Update {A->step[ax]}: */
        if (st[i] != 0)
          { if (skip > 0)
              { /* Update {*bp} (can't overflow because element exists): */
                (*bp) += skip * st[i];
              }
            /* If index is now trivial, clear step: */
            if (keep == 1) { st[i] = 0; }
          } 
      } 
  }

void ix_subsample 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_size_t stride 
  )
  { /* Check axis: */
    demand(i < d, "bad axis");
    /* Check stride: */
    demand(stride > 0, "bad stride");
    /* Check for trivial subsampling: */
    if (stride == 1) { return; }
    /* Save the original size along axis {i}: */
    ix_size_t osz = sz[i];
    /* If array is empty or trivial along that index, there is nothing to do: */
    if (osz <= 1) { return; }
    /* Compute and store the new size: */
    ix_size_t nsz = (osz-1)/stride + 1;
    assert(nsz > 0);
    sz[i] = nsz;
    /* Adjust step if needed: */
    if (st[i] != 0)
      { if (nsz == 1) 
          { /* Index {i} became trivial: */
            st[i] = 0;
          }
        else
          { /* Index {i} now is {stride} times as powerful: */
            st[i] *= stride;
          }
      }
  }

void ix_flip 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i
  )
  { /* Check axis: */
    demand(i < d, "bad axis");
    /* If there is no motion along that axis, we have nothing to d: */
    ix_step_t sti = st[i];
    if (sti == 0) { return; }
    /* Now {sz[i]} must be positive. */
    /* Adjust {*bp} to the last valid index value: */
    (*bp) += (sz[i]-1)*sti; 
    /* Negate the step {st[i]}: */
    st[i] = -sti;
  }

void ix_swap_indices 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_axis_t j 
  )
  { /* Check axes: */
    demand(i < d, "bad axis i");
    demand(j < d, "bad axis j");
    /* Trivial transpose: */
    if (i == j) { return; }
    /* Exchange parameters of axes {i,j}: */
    { ix_size_t tmp = sz[i]; sz[i] = sz[j]; sz[j] = tmp; }
    { ix_pos_t  tmp = st[i]; st[i] = st[j]; st[j] = tmp; }
  }

void ix_flip_indices
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_axis_t j 
  )
  { /* Check axes: */
    demand(i < d, "bad axis i");
    demand(j < d, "bad axis j");
    /* Swap symmetric entries: */
    while (i < j)
      { /* Exchange parameters of axes {i,j}: */
        { ix_size_t tmp = sz[i]; sz[i] = sz[j]; sz[j] = tmp; }
        { ix_pos_t tmp = st[i]; st[i] = st[j]; st[j] = tmp; } 
        i++; j--;
      }
  }

void ix_replicate 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_size_t size 
  )
  { /* Check axis: */
    demand(i < d, "bad axis");
    /* Axis {i} must be trivial: */
    demand(sz[i] == 1, "index is not trivial");
    /* Check and set the new size: */
    demand((size > 0) && (size <= ix_MAX_SIZE), "bad new size");
    sz[i] = size;
    /* The increment must already be zero, but just in case... */
    st[i] = 0;
  }

void ix_diagonal 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i,
    ix_axis_t j 
  )
  { /* Check axes: */
    demand(i < d, "bad axis i");
    demand(j < d, "bad axis j");
    demand(i != j, "axes are equal");
    /* Require the array to be wide enough: */
    demand(sz[j] >= sz[i], "array has wrong shape");
    /* Require the first index to be non-empty: */
    demand(sz[i] > 0, "array is empty on the first index");
    /* The step along {i} is the sum of both steps: */
    if (sz[i] > 1) { st[i] += st[j]; }
    /* The size along {j} is reduced by {sz[i]-1}: */
    sz[j] = sz[j] - (sz[i] - 1); 
    /* May become trivial (but not empty) along {j}: */
    if (sz[j] <= 1) { st[j] = 0; }
  }

void ix_chop 
  ( ix_dim_t d, 
    ix_size_t sz[], 
    ix_pos_t *bp, 
    ix_step_t st[], 
    ix_axis_t i, 
    ix_size_t stride, 
    ix_axis_t j 
  )
  { /* Check axes: */
    demand(i < d, "bad axis i");
    demand(j < d, "bad axis j");
    demand(i != j, "axes must be different");
    /* The chunk size must be positive: */
    demand(stride > 0, "bad chunk size");
    /* The array must be trivial in direction {j}: */
    demand(sz[j] == 1, "axis j not trivial");
    /* Save the original size along axis {i}: */
    ix_size_t osz = sz[i];
    /* The size must be a multiple of the chunk size: */
    demand(osz % stride == 0, "non-integral number of chunks");
    /* Compute new sizes along {j} and {i}: */
    sz[j] = osz/stride;  sz[i] = stride;
    /* Compute new steps along {j} and {i}: */
    if (sz[j] == 1)
      { /* Array is still trivial along {j}; just in case: */
        st[j] = 0;
      }
    else
      { /* This computation cannot overflow.: */
        st[j] = stride * st[i];
      }
    /* Check whether axis {i} became trivial: */
    if (stride == 1) { st[i] = 0; } 
  }

/* INDEX TUPLE STEPPING */

bool_t ix_next_C ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[], ix_step_t st[], ix_pos_t *p )
  { /* Look for an index {ix[i]} that can be incremented: */
    int i = d-1;
    ix_index_t *ixi = ix + i;
    while ((*ixi) >= sz[i]-1) 
      { /* Index {ix[i]} is at its limit, reset to 0: */
        if (p != NULL) { (*p) -= (*ixi)*st[i]; }
        (*ixi) = 0;
        /* Wrap around? */
        if (i <= 0) { return TRUE; }
        /* Go to next index: */
        i--; ixi--;
      }
    /* Increment index {ix[i]} */
    (*ixi)++; 
    if (p != NULL) { (*p) += st[i]; }
    return FALSE;
  }
    
bool_t ix_next_F ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[], ix_step_t st[], ix_pos_t *p )
  { /* Look for an index {ix[i]} that can be incremented: */
    int i = 0;
    ix_index_t *ixi = ix;
    while ((*ixi) >= sz[i]-1) 
      { /* Index {ix[i]} is at its limit, reset to 0: */
        if (p != NULL) { (*p) -= (*ixi)*st[i]; }
        (*ixi) = 0;
        /* Go to next index: */
        i++;
        /* Wrap around? */
        if (i >= d) { return TRUE; }
        ixi++;
      }
    /* Increment index {ix[i]} */
    (*ixi)++; 
    if (p != NULL) { (*p) += st[i]; }
    return FALSE;
  }
    
bool_t ix_prev_C ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[], ix_step_t st[], ix_pos_t *p )
  { /* Look for an index {ix[i]} that can be decremented: */
    int i = d-1;
    ix_index_t *ixi = ix + i;
    while ((*ixi) <= 0) 
      { /* Index {ix[i]} is at its limit, reset to {sz[i]-1}: */
        (*ixi) = sz[i]-1;
        if (p != NULL) { (*p) += (*ixi)*st[i]; }
        /* Wrap around? */
        if (i <= 0) { return TRUE; }
        /* Go to next index: */
        i--; ixi--;
      }
    /* Decrement index {ix[i]} */
    (*ixi)--;
    if (p != NULL) { (*p) -= st[i]; }
    return FALSE;
  }

bool_t ix_prev_F ( ix_dim_t d, ix_index_t ix[], ix_size_t sz[], ix_step_t st[], ix_pos_t *p )
  { /* Look for an index {ix[i]} that can be decremented: */
    int i = 0;
    ix_index_t *ixi = ix;
    while ((*ixi) <= 0) 
      { /* Index {ix[i]} is at its limit, reset to {sz[i]-1}: */
        (*ixi) = sz[i]-1;
        if (p != NULL) { (*p) += (*ixi)*st[i]; }
        /* Go to next index: */
        i++;
        /* Wrap around? */
        if (i >= d) { return TRUE; }
        ixi++;
      }
    /* Decrement index {ix[i]} */
    (*ixi)--;
    if (p != NULL) { (*p) -= st[i]; }
    return FALSE;
  }
  
sign_t ix_compare_C ( ix_dim_t d, ix_index_t ixa[], ix_index_t ixb[] )
  { int i;
    ix_index_t *ixai = ixa, *ixbi = ixb;
    for (i = 0; i < d; i++)
      { if ((*ixai) < (*ixbi))
          { return NEG; }
        else
          { return POS; }
        ixai++; ixbi++;
      }
    return ZER;
  }
  
sign_t ix_compare_F ( ix_dim_t d, ix_index_t ixa[], ix_index_t ixb[] )
  { int i;
    ix_index_t *ixai = ixa+d, *ixbi = ixb+d;
    for (i = d-1; i >= 0; i--)
      {  ixai--; ixbi--;
        if ((*ixai) < (*ixbi))
          { return NEG; }
        else
          { return POS; }
      }
    return ZER;
  }
    
/* ELEMENT ENUMERATION */

void ix_enum 
  ( ix_op_t op,
    ix_dim_t d,
    ix_size_t sz[], 
    ix_pos_t bpA, 
    ix_step_t stA[], 
    ix_pos_t bpB, 
    ix_step_t stB[], 
    ix_pos_t bpC, 
    ix_step_t stC[]
  )
  { ix_pos_t pA = bpA;
    ix_pos_t pB = bpB;
    ix_pos_t pC = bpC;
    
    ix_index_t ix[d];
    /* Initialize index tuple: */
    { int i;
      for (i = 0; i < d; i++) 
        { if (sz[i] <= 0) { /* Oops, array is empty: */ return; }
          ix[i] = 0;
        }
    }
    /* Main loop: */
    while(1)
      { /* Operate on element: */
        op(ix, pA, pB, pC);
        /* Look for an index {ix[i]} that can be incremented: */
        int i = d-1;
        ix_index_t *ixi = ix + i;
        while (*ixi >= sz[i]-1) 
          { /* Index {ix[i]} is at its limit: */
            if ((*ixi) > 0)
              { /* Reset index to 0, adjust positions: */
                if (stA != NULL) pA -= (*ixi)*stA[i];
                if (stB != NULL) pB -= (*ixi)*stB[i]; 
                if (stC != NULL) pC -= (*ixi)*stC[i];
                (*ixi) = 0;
              }
            if (i <= 0) { /* Exhausted all index tuples: */
               return; }
            /* Go to next index: */
            i--; ixi--;
          }
        /* Increment index {ix[i]} */
        (*ixi)++;
        /* Advance positions: */
        if (stA != NULL) pA += stA[i];
        if (stB != NULL) pB += stB[i]; 
        if (stC != NULL) pC += stC[i];
      }
 }

/* PARAMETER VALIDATION */

bool_t ix_parms_are_valid ( ix_dim_t d, ix_size_t sz[], ix_pos_t bp, ix_step_t st[], bool_t die )
  { /* Check the position of element {[0,0,...,0]}: */
    if (bp > ix_MAX_POS) { fail_test(die,"bp too big"); }
    
    /* Check sizes and number of (supposedly) distinct positions: */
    ix_dim_t dw = 0; /* Number of non-trivial indices. */
    ix_elem_count_t npos = 1; /* Number of index tuples, excl. replications. */
    ix_axis_t i;
    for (i = 0; i < d; i++) 
      { ix_step_t sti = st[i];
        ix_size_t szi = sz[i];
        /* Count number of axes with non-trivial steps: */
        if (sti != 0) { dw++; }
        /* Validate size and step: */
        if (szi == 0) 
          { /* Array is empty: */ npos = 0; }
        else if (szi >= 2)
          { /* Check size range: */
            if (szi > ix_MAX_SIZE) { fail_test(die,"array size too big"); }
            if (sti > 0)
              { /* Axis is not replicated; more positions (worry about overflow): */
                if (szi > ix_MAX_POSITIONS / npos) { fail_test(die,"too many positions"); }
                npos *= szi;
              }
          }
      }
    if (npos == 0)
      { /* Empty array -- all steps and the base must be zero: */
        if (bp != 0)  { fail_test(die,"nonzero base in empty array"); }
        if (dw != 0)  { fail_test(die,"nonzero step in empty array"); }
        return TRUE;
      }
    /* Check number of (supposedly) distinct positions: */
    if (npos > ix_MAX_POSITIONS) { fail_test(die,"too many positions"); }

    /* Check steps and min/max position */
    ix_pos_t lop = bp; /* Minimum position. */
    ix_pos_t phi = bp; /* Maximum position. */
    for (i = 0; i < d; i++) 
      { ix_step_t sti = st[i];
        ix_size_t szi = sz[i];
        /* Validade size and step: */
        assert(szi > 0); /* Else we should have returned already. */
        if (szi == 1)
          { /* Axis is trivial, step must be zero: */
            if (sti != 0) { fail_test(die,"nonzero step in trivial axis"); }
          }
        else if (sti > 0)
          { /* Check step position: */
            if (sti > ix_MAX_ABS_STEP) { fail_test(die,"step too big"); }
            /* Check that the maximum position is within range: */
            if ((ix_MAX_POS - phi)/(szi - 1) < sti) { fail_test(die,"position too big"); }
            phi += (szi - 1)*sti;
          }
        else if (sti < 0)
          { /* Beware that we can't make {sti} positive before checking: */
            if (sti < -ix_MAX_ABS_STEP) { fail_test(die,"step too big"); }
            /* Make step positive: */
            sti = -sti;
            /* Check that minimum position is non-negative: */
            if (lop/(szi - 1) < sti) { fail_test(die,"negative position"); }
            lop -= (szi - 1)*sti;
          }
        else
          { /* Step is zero; axis is replicated, ignore: */ }
      }
    /* Well, everything seems allright... */
    return TRUE; 
  }

bool_t ix_positions_are_distinct ( ix_dim_t d, ix_size_t sz[], ix_step_t st[], bool_t die )
  { /* Check {sz,st}: */

    /* The injectivity is hard to check exactly, so this procedure
    checks first whether there is a sequence of transpositions and
    flips that makes the lexicographical order of the index tuples
    (ignoring replicated axes) imply strict ordering of the positions.
    
    More precisely, let {stw[0..dw-1]} be the absolute values of the
    nonzero steps, sorted in increasing order, and {szw[0..dw-1]} the
    corresponding sizes. A sufficient condiiton for injectivity is that
      { stw[i] > SUM { stw[j]*(szw[j]-1) : j \in 0..i-1 } }
    for all {i} in {0..dw-1}. (This means that the displacement due to
    an increment of 1 in the index corresponding to {stw[i]} exceeds
    the maximum displacement that can be obtained with all the
    previous indices combined.) */

    ix_step_t stw[d]; /* Nonzero steps in increasing order */
    ix_size_t szw[d]; /* Corresponding sizes in the same order */
    ix_dim_t dw = 0; /* Number of non-trivial indices. */
    ix_axis_t i;
    for (i = 0; i < d; i++) 
      { ix_step_t sti = st[i];
        ix_size_t szi = sz[i];
        if (sti != 0)
          { /* Assume the sizes are consistent with the steps: */
            assert(szi >= 2); 
            /* Assume {sti} is not too big to negate: */
            if (sti < 0) { sti = -sti; }
            /* Save {szi,sti} in increasing order: */
            int j = dw;
            while ((j > 0) && (stw[j-1] > sti))
              { stw[j] = stw[j-1];
                szw[j] = szw[j-1];
                j--;
              }
            stw[j] = sti;
            szw[j] = szi;
            dw++;
          }
      }
    /* Compute {tsw[i]} = max displ. using indices {0..i}. */
    ix_pos_t tdw[d]; 
    ix_pos_t tdisp = 0; /*  */
    for (i = 0; i < dw; i++)
      { tdisp += stw[i]*(szw[i]-1);
        tdw[i] = tdisp;
      }
    /* Exclude trailing indices which are strictly monotonic: */
    while ((dw > 0) && (stw[dw-1] > tdw[dw-1])) { dw--; }
    /* If no indices are left, the positions are strictly monotonic: */ 
    if (dw == 0) { return TRUE; }
   
    /* Positions are not motonic, rats. Must do the hard test. */
    /* There should be a smart way to do it. We use brute force: */
    /* we count the number {ct[p]} of index tuples per position {p}. */
    ix_elem_count_t npos = tdw[dw-1] + 1;
    uint8_t *ct = (uint8_t *)malloc(npos*sizeof(uint8_t)); 
    affirm(ct != NULL, "no memory for reachability counters");
    ix_pos_t p;
    /* Clear all counters: */
    for (p = 0; p < npos; p++) { ct[p] = 0; }
    /* Enumerate index tuples and bump counters: */
    ix_index_t ixw[dw];
    for (i = 0; i < dw; i++) { ixw[i] = 0; }
    p = 0;
    bool_t collision = FALSE;
    do 
      { assert((p >= 0) && (p < npos));
        if (ct[p] > 0) { collision = TRUE; break; }
        ct[p]++;
      } 
    while(! ix_next_C(dw,ixw,szw,stw,&p));
    /* Free storage before returning anything: */
    free(ct); 
    /* Verdict: */
    if (! collision) { return TRUE; }
    fail_test(die,"positions are not distinct");
  }

