/* vec.h -- self-bounded vectors (one-dimensional arrays) of things */
/* Last edited on 2007-02-05 18:50:06 by stolfi */

/* This module defines tools for handling generic vectors
  (unidimensional arrays) of things, where, unlike standard C vectors,
  the number of elements is stored as part of the representation.
  Thus, when passing a {vec_t} to a procedure, there is no need to
  pass its size as as a separate parameter.
  
  This module also defines efficient and convenient procedures for
  extending and trimming vectors as needed. Expansion doubles the size
  of the vector, so the total cost of the expansion is small and
  proportional to the final size of the vector.
  
  In many contexts, these vectors can replace linked lists, with
  advantages. A typical example would be
  
    double_vec_t x = double_vec_new(initsize);
    { int nx = 0;
      for ...
        { ...
          double_vec_expand(&x, nx);
          x.el[nx] = ...; nx++;
        }
      double_vec_trim(&x, nx);
    }
    for (i = 0; i < x.nel; i++) { print(x.el[i]); }

  */

#ifndef vec_H
#define vec_H

#include <stdlib.h>
#include <bool.h>
#include <ref.h>
#include <stdint.h>

/* TYPED VECTOR DESCRIPTORS
  
  To define a new vector type called, e.g., {myvec_t}, with elements 
  of type {myelem_t}, write 
  
    { vec_typedef(myvec_t,myvec,myelem_t); }
    
  in the proper ".h" file, and 
  
    { vec_typeimpl(myvec_t,myvec,myelem_t); }
    
  in the corresponding ".c" file. These macros will declare the type
  {myvec_t} and the four associated procedures:
  
    { myvec_t myvec_new(unsigned int nel); }
    
    { void myvec_expand(myvec_t *v, unsigned int index); }
    
    { void myvec_trim(myvec_t *v, unsigned int size); }

    { myvec_t myvec_desc(unsigned int nel, myelem_t *el); }
    
  A variable {v} of type {myvec_t} is a descriptor to a 
  vector of {myelem_t} elements.  The field {v.el} points to a heap
  area with space for {v.nel} such elements, namely {v.el[0..v.nel-1]}. 

  The call {myvec_new(nel)} allocates a new {myvec_t} with
  space for {nel} elements, and returns its {myvec_t} descriptor.

  The call {myvec_expand(&v,index)} makes sure that element
  {v.el[index]} exists. If necessary, it reallocates the area
  referenced by {v.el}, copies the old elements into it, and updates
  {v.nel} to the new size (which will be strictly greater than
  {index}). Size is doubled at each reallocation, to ensure total
  {O(N)} time for {N} calls.

  The call {myvec_trim(&v,size)} reallocates {v.el} if necessary
  so that it has exactly {size} elements. It preserves the elements
  {v.el[0..size-1]} and sets {v.nel = size}.
  
  The call {myvec_desc(el, nel)} assembles a {myvec_t} descriptor from
  a given element count {nel} and the address {el} of the first
  element. The client must ensure that {el[0..nel-1]} are valid. The
  procedures {myvec_expand} and {myvec_trim} can be applied to the
  resulting descriptor only if the address {el} was provided by
  {malloc}.

  The macros {vec_typedef} and {vec_typeimpl} are defined with three
  separate parameters for greater flexibility and back-compatibility.
  However, it is suggested that the parameters be of the form
  {XXX_vec_t}, {XXX_vec}, and {XXX_t}, for some {XXX}, whenever
  possible.
  
*/

#define vec_typedef(VECTYPE,VECNAME,ELEMTYPE) \
typedef struct VECTYPE { unsigned int nel; ELEMTYPE *el; } VECTYPE; \
VECTYPE VECNAME##_new(unsigned int nel); \
void VECNAME##_expand(VECTYPE *nv, int index); \
void VECNAME##_trim(VECTYPE *nv, unsigned int nel); \
VECTYPE VECNAME##_desc(ELEMTYPE *el, unsigned int nel)

#define vec_typeimpl(VECTYPE,VECNAME,ELEMTYPE) \
VECTYPE VECNAME##_new(unsigned int nel) \
  { vec_t v = vec_new(nel, sizeof(ELEMTYPE)); \
    return (VECTYPE){v.nel, (ELEMTYPE *)v.el}; \
  } \
void VECNAME##_expand(VECTYPE *nv, int index) \
  { vec_expand(vec_cast_ref(nv), index, sizeof(ELEMTYPE)); } \
void VECNAME##_trim(VECTYPE *nv, unsigned int nel) \
  { vec_trim(vec_cast_ref(nv), nel, sizeof(ELEMTYPE)); } \
VECTYPE VECNAME##_desc(ELEMTYPE *el, unsigned int nel) \
  { return (VECTYPE){nel, el}; } \
extern void VECNAME##_bOgUs /* To eat the semicolon. */

/* GENERIC VECTOR DESCRIPTORS */

typedef struct vec_t /* General vector descriptor. */
  { unsigned int nel;   /* Number of elements. */
    void *el;  /* Pointer to element zero. */
  } vec_t;
  /* A {vec_t} {v} describes a uni-dimensional array of {v.nel} elements
    of the same size and type, stored in consecutive memory locations,
    starting at address {v.el}.
    
    The size of each element is not part of the structure, and must be
    provided separately. Usually, {vec_t}s are used to implement
    vectors of a specific type, such as {int} or {unsigned int}, where the size
    can be provided by macros. See the examples of {int_vec_t} and
    {char_vec_t} below. */

/* GENERIC VECTOR FUNCTIONS 
  
  The functions in this section may be called directly by clients,
  but their main purpose is to implement the 
  functions {myvec_new}, {myvec_expand}, and {myvec_trim}
  defined by {vec_typedef} and {vec_typeimpl}. */

vec_t vec_new(int nel, size_t elsz);
  /* Allocates a new vector with {nel} elements of size {elsz}.
    Bombs out if there is no space for the request. 
    If {nel == 0}, the result has {el == NULL}. */

void vec_expand(vec_t *v, unsigned int index, size_t elsz);
  /* Makes sure that element {v->el[index]} exists, reallocating and
    copying the array {*v->el} if {index >= v->nel}. If that happens,
    the new {v->nel} will be about twice as big as the old one. */

void vec_trim(vec_t *v, unsigned int nel, size_t elsz);
  /* Makes sure that {v.nel == nel}, reallocating and copying 
    the  array {*v->el} if {v.nel != nel}. If {nel == 0}, the 
    result will have {el == NULL}. */

vec_t *vec_cast_ref(void *v);
  /* Casts the argument as an address to a {vec_t}. Can be used to
    apply vec_expand or vec_trim to vectors of specific types. 
    We define it as a function in order to trigger type warnings 
    when {v} is not an address (a common mistake). */

/* SUB-VECTOR DESCRIPTORS */

typedef struct vec_sub_t /* Regular sub-sequence of a {vec_t}. */
  { unsigned int nel;   /* Number of elements in sub-vector. */
    void *el;  /* Pointer to element zero of sub-vector. */
    int step;  /* Increment between consecutive elements (in elements). */
  } vec_sub_t;

vec_sub_t vec_as_sub(vec_t *v);
  /* Returns a sub-vector descriptor for the entire vector
    {v} (with {step == 1}). */
    
vec_sub_t vec_sub(vec_sub_t *vs, unsigned int start, int step, unsigned int nel, size_t elsz);
  /* Returns a descriptor {r} for the specified subsequence of elements
    of {vs}, namely {r[i] = vs[start + i*step]}, for {i = 0..r.nel-1}.
    The {step} cannot be zero.
    
    The result size {r.nel} is the largest integer in {0..nel} such
    that all those elements exist. In particular, if {start} is not in
    {0..vs.nel-1}, then {m} is zero. */
    
/* SOME USEFUL TYPED VECTORS */

vec_typedef(int_vec_t,    int_vec,    int);
  /* Vectors of {int}s. */

vec_typedef(uint_vec_t,   uint_vec,   unsigned int);
  /* Vectors of {unsigned int}s. */

vec_typedef(float_vec_t,  float_vec,  float);
  /* Vectors of {float}s */

vec_typedef(double_vec_t, double_vec, double);
  /* Vectors of {double}s. */

vec_typedef(bool_vec_t,   bool_vec,   bool_t);
  /* Vectors of {bool_t}s. */

vec_typedef(char_vec_t,   char_vec,   char);
  /* Vectors of {char}s. */

vec_typedef(ubyte_vec_t,   ubyte_vec, uint8_t);
  /* Vectors of {unsigned char}s. */

vec_typedef(sbyte_vec_t,   sbyte_vec, int8_t);
  /* Vectors of {signed char}s. */

vec_typedef(string_vec_t, string_vec, char*);
  /* Vectors of strings ({char*}s). */

vec_typedef(ref_vec_t, ref_vec, ref_t);
  /* Vectors of arbitrary addresses. */

#endif
