/* See pscolor.h */
/* Last edited on 2005-06-05 15:16:09 by stolfi */

#include <pswr_color.h>
#include <pswr_iso.h>
#include <affirm.h>
#include <math.h>
#include <stdlib.h>
#include <values.h>

void pswr_make_color_table
  ( double vStart,
    double vStep,
    int kMin,
    int kMax, 
    double RMin, double GMin, double BMin,
    double RZer, double GZer, double BZer,
    double RMax, double GMax, double BMax,
    int *NP, 
    double **RP, 
    double **GP,
    double **BP
  )
  {
    /* Create gradual color table: */
    int N = kMax - kMin + 2;     /* Number of color bands: */
    double *R = (double *)malloc(N*sizeof(double));
    double *G = (double *)malloc(N*sizeof(double));
    double *B = (double *)malloc(N*sizeof(double));
    double vMin = vStart + (kMin - 0.5)*vStep;
    double vMax = vStart + (kMax + 0.5)*vStep;
    
    int i;
    for (i = 0; i < N; i++)
      { int k = kMin + i;  /* Index of upper isoline of band {i}. */
        
        /* Compute relative function value {r} at mid-band, in [-1 _ +1]: */
        double r; 
        if (i == 0) 
          { r = -1.0; }
        else if (i == N-1) 
          { r = +1.0; }
        else
          { /* Function values at lower and upper isolines of band: */
            double vLo = pswr_level(vStart, vStep, k-1);
            double vHi = pswr_level(vStart, vStep, k);
            /* Compute nominal function value {vMd} at mid-band: */
            double vMd = (vLo + vHi)/2;
            affirm ((vMin < vMd) && (vMd < vMax), "bug");
            r = (vMd > 0.0 ? vMd/vMax : -vMd/vMin);
          }
          
        /* Intepolate between colors in equal perceptual distance: */
        double *Ri = &(R[i]), *Gi = &(G[i]), *Bi = &(B[i]);
        if (r > 0.0)
          { pswr_interpolate_colors(+r, RZer,GZer,BZer, RMax,GMax,BMax, Ri,Gi,Bi); }
        else
          { pswr_interpolate_colors(-r, RZer,GZer,BZer, RMin,GMin,BMin, Ri,Gi,Bi); }
      }
      
    *NP = N;
    *RP = R; *GP = G; *BP = B;
  }

void pswr_interpolate_colors
  ( double r, 
    double R0, double G0, double B0,
    double R1, double G1, double B1,
    double *R, double *G, double *B
  )
  {
    /* Compute brightnesses {Y0,Y1} of {R0,G0,B0} and {R1,G1,B1}: */ 
    double Y0 = 0.299*R0 + 0.587*G0 + 0.114*B0;
    double Y1 = 0.299*R1 + 0.587*G1 + 0.114*B1;
    /* Interpolate brightness {Y} in cubic-root scale: */
    double w = (1-r)*cbrt(Y0) + r*cbrt(Y1);
    double Y = w*w*w;
    if (Y < 0) { Y = 0; }
    if (Y > 1) { Y = 1; }
    
    /* Compute limiting chroma vectors: */ 
    double dR0 = R0 - Y0, dG0 = G0 - Y0, dB0 = B0 - Y0;
    double dR1 = R1 - Y1, dG1 = G1 - Y1, dB1 = B1 - Y1;
    /* Interpolate chroma vector {dR,dG,dB} in linear scale: */
    double dR = (1-r)*dR0 + r*dR1;
    double dG = (1-r)*dG0 + r*dG1;
    double dB = (1-r)*dB0 + r*dB1;
    
    /* Compute max {h <= 1} h.t. {(Y,Y,Y) + h*(dR,dG,dB)} is ok: */
    double h = 1.0, lim;
    lim = (dR < 0 ? 0 : 1) - Y; if (h*dR*dR > lim*dR) { h = lim/dR; }
    lim = (dG < 0 ? 0 : 1) - Y; if (h*dG*dG > lim*dG) { h = lim/dG; }
    lim = (dB < 0 ? 0 : 1) - Y; if (h*dB*dB > lim*dB) { h = lim/dB; }
    /* Now add the clipped chroma component: */
    (*R) = Y + h*dR;
    (*G) = Y + h*dG;
    (*B) = Y + h*dB;
  }

void pswr_color_scale_1
  ( double f, 
    double R0, double G0, double B0,
    double *R, double *G, double *B
  )
  {
    /* Compute brightness {Y0} of {R0,G0,B0}: */ 
    double Y0 = 0.299*R0 + 0.587*G0 + 0.114*B0;
    /* Compute limiting chroma vector (for {m = 1}): */ 
    double dR = R0 - Y0, dG = G0 - Y0, dB = B0 - Y0;
    if (f < 0) { dR = -dR; dG = -dG; dB = -dB; }
    /* Compute brightness {Y} of target color: */ 
    double m = fabs(f);
    double bias = 0.125;
    double Y0b = (Y0 + bias)/(1 + bias);
    double Y = (1 + bias)*exp(m*log(Y0b)) - bias;
    if (Y < 0) { Y = 0; }
    /* Compute max {h <= 1} h.t. {(Y,Y,Y) + h*(dR,dG,dB)} is ok: */
    double h = 1.0, lim;
    lim = (dR < 0 ? 0 : 1) - Y; if (h*dR*dR > lim*dR) { h = lim/dR; }
    lim = (dG < 0 ? 0 : 1) - Y; if (h*dG*dG > lim*dG) { h = lim/dG; }
    lim = (dB < 0 ? 0 : 1) - Y; if (h*dB*dB > lim*dB) { h = lim/dB; }
    /* Now add chroma component: */
    m *= h;
    (*R) = Y + m*dR;
    (*G) = Y + m*dG;
    (*B) = Y + m*dB;
  }

void pswr_color_scale_2
  ( double fs, double ft, 
    double R0, double G0, double B0,
    double *R, double *G, double *B
  )
  {
    /* Convert {R0,G0,B0} to YUV coordinates: */ 
    double Y0 = 0.299*R0 + 0.587*G0 + 0.114*B0;
    double U0 = 0.284*(R0 - G0);
    double V0 = - 0.139*R0 - 0.085*G0 + 0.224*B0;
    /* Compute zero-brightness component of {R0,G0,B0}: */ 
    double sR = R0 - Y0, sG = G0 - Y0, sB = B0 - Y0;
    /* Compute zero-brightness vector orthogonal to {sR,sG,sB}: */ 
    double U1 = -V0, V1 = U0;
    double tR = + 2.2214430220*U1 - 0.50943800940*V1;
    double tG = - 1.2996837380*U1 - 0.50943800940*V1;
    double tB = + 0.8853011714*U1 + 3.95484770500*V1;
    /* Compute brightness {Y} of target color: */ 
    double m = hypot(fs, ft);
    double bias = 0.125;
    double Y0b = (Y0 + bias)/(1 + bias);
    double Y = (1 + bias)*exp(m*log(Y0b)) - bias;
    if (Y < 0) { Y = 0; }
    if (Y > 1) { Y = 1; }
    /* Compute limiting chroma vector (for {m = 1}): */
    double dR = fs/m * sR + ft/m * tR;
    double dG = fs/m * sG + ft/m * tG;
    double dB = fs/m * sB + ft/m * tB;
    /* Compute max {h <= 1} s.t. {(Y,Y,Y) + h*(dR,dG,dB)} is ok: */
    double h = 1.0, lim;
    lim = (dR < 0 ? 0 : 1) - Y; if (h*dR*dR > lim*dR) { h = lim/dR; }
    lim = (dG < 0 ? 0 : 1) - Y; if (h*dG*dG > lim*dG) { h = lim/dG; }
    lim = (dB < 0 ? 0 : 1) - Y; if (h*dB*dB > lim*dB) { h = lim/dB; }
    /* Now add chroma component: */
    m *= h;
    (*R) = Y + m*dR;
    (*G) = Y + m*dG;
    (*B) = Y + m*dB;
  }
  
void pswr_color_scale_3
  ( double fu, double fv, double fw, 
    double *R, double *G, double *B
  )
  {
    /* Choose three equally spaced zero-brightness vectors: */
    double uR = +0.63089, uG = -0.36911, uB = +0.25143;
    double vR = -0.44074, vG = +0.05926, vB = +0.84696;
    double wR = -0.19015, wG = +0.30985, wB = -1.09838;
    /* Compute brightness {Y} of target color: */ 
    double fm = (fu+fv+fw)/3.0;
    double Ymin = 0.125;
    double bias = 0.125;
    double Yminb = (Ymin + bias)/(1 + bias);
    double Y = (Ymin + bias)*exp(fm*log(Yminb)) - bias;
    if (Y < 0) { Y = 0; }
    if (Y > 1) { Y = 1; }
    /* Remove mean value from {fu,fv,fw}: */
    fu -= fm; fv -= fm; fw -= fm;
    /* Compute min {fh} s.t. {(fm,fm,fm)+(fu,fv,fw)/fh} is not OK: */
    double fh;
    if ((fm > 0.9999) || (fm < 0.0001))
      { fh = 1.0; }
    else
      { double L;
        fh = 0.0;
        L = (fu < 0 ? 0 : 1) - fm; if (fu*fu > fh*L*fu) { fh = fu/L; }
        L = (fv < 0 ? 0 : 1) - fm; if (fv*fv > fh*L*fv) { fh = fv/L; }
        L = (fw < 0 ? 0 : 1) - fm; if (fw*fw > fh*L*fw) { fh = fw/L; }
      }
    /* Compute chroma direction vector: */
    double dR = fu * uR + fv * vR + fw * wR;
    double dG = fu * uG + fv * vG + fw * wG;
    double dB = fu * uB + fv * vB + fw * wB;
    /* Compute max {dh} s.t. {(Y,Y,Y) + dh*(dR,dG,dB)} is OK: */
    double dh = MAXDOUBLE;
    { double L;
      L = (dR < 0 ? 0 : 1) - Y; if (dh*dR*dR > L*dR) { dh = L/dR; }
      L = (dG < 0 ? 0 : 1) - Y; if (dh*dG*dG > L*dG) { dh = L/dG; }
      L = (dB < 0 ? 0 : 1) - Y; if (dh*dB*dB > L*dB) { dh = L/dB; }
    }
    /* Now add chroma component: */
    double m = dh*fh; 
    (*R) = Y + m*dR;
    (*G) = Y + m*dG;
    (*B) = Y + m*dB;
  }

