#ifndef oct_enum_H
#define oct_enum_H

/* Element enumeration procedures for the general quad-edge structure. */
/* Last edited on 2007-02-08 17:48:20 by stolfi */

#define oct_enum_H_copyright \
  "Copyright  1996, 2006 Institute of Computing, Unicamp."

#include <oct.h>

/* ENUMERATION PROCEDURES
  
  For this interface, an /enumeration procedure/ is a function that
  locates all the arcs of the quad-edge structure that are reachable
  from a given list of /root arcs/ by zero or more /steps/, in all
  combinations; and /visits/ some of those arcs.
  
  Visiting an arc means passing it to caller for arbitrary processing
  (see {VISIT-FUNCTIONS} below) and/or saving it in an {arcs_t} array
  provided by the caller (see {VISITATION LISTS} below).

  An enumeration procedure never visits the same arc more than
  once, but may visit only a proper subset of the arcs it found ---
  e.g. only one arc on each vertex or face of the map.

  The enumeration procedures below do not modify the data structure or
  any global storage (except possibly through the the client's
  visit-func). Therefore, concurrent calls (e.g. in multi-threaded
  programming or enumeration-within-enumeration) do not
  directly interfere with each other. */
  
/* STEP-FUNCTIONS

  Each step of the enumeration consists of following some /step-func/,
  a function that maps arcs to arcs --- such as {onext}, {sym},
  {fflip} etc.. 
  
  Some of the procedures below, such as {enum_ring} and
  {enum_orbits}, use arbitrary step-funcs provided by the caller.
  Other procedures use a fixed set of step-funcs, defined internally. */

typedef oct_arc_t (*oct_step_t)(oct_arc_t a);
 /* A function from items to items. */

/* VISIT-FUNCTIONS

  The enumeration procedures below generally take the address of a
  /visit-func/ as a parameter (of type {oct_visit_t}) to process each
  arc found in the enumeration.

  The visit-func may return TRUE to abort the enumeration, or FALSE to
  let it continue. Unless noted otherwise, the enumeration procedure
  itself will return TRUE iff any visit-func call returned TRUE.
  
  The visit-func should not try to abort the enumeration with a long
  jump, since doing so will create a storage leak and may even leave
  the structure in an invalid state.

  A {NULL} visit-func is treated as a no-op that always returns
  {FALSE}. */

typedef bool_t (*oct_visit_t)(oct_arc_t p);
 /* A client-provided visitation procedure. */

/* VISITATION LISTS

  An enumeration procedure also takes a parameter {vP} of type
  {oct_arcs_t*}. If {vP} is not {NULL}, the procedure will store into
  {vP} each item that it visits, in order of visitation.

  Typically, a client that uses this feature will set {vP} to the
  address of an empty {oct_arcs_t}, with {vP->el == NULL} and {vP->nel
  == 0}. (Note that this is different from passing {vP == NULL}.) In
  any case, the new entries will be inserted *after* any entries
  that are already in {*vP}, namely starting at position {vP->el[vP->nel]}.

  Visited arcs are stored in {*vP} even if the visit-func is {NULL},
  or if the enumeration is aborted by a visit-func returning TRUE. On
  the other hand, the procedure will not store in {*vP} any
  intermediate arcs that were used to reach the visited ones but were
  not themselves visited.

  The count {vP->nel} is updated to the number of arcs visited. The
  storage area {*(vP->el)} is expanded as needed, and trimmed at
  the end of the enumeration to hold precisely {vP->nel} arcs. */

/* VIEW-PRESERVING ENUMERATIONS

  A step-func {stp} is /view-preserving/ if it preserves the 
  dual/primal character of the argument; that is, {a.M == a.stp.M} 
  for every arc {a}. In that case, every arc {a} found in the enumeration
  satisfies {a.M == a.M} for some root arc {r}. Thus, if all roots
  are on the same map, only arcs on that map will be visited. Of
  course, the arcs on the other map can be obtained by applying
  `{*}' to the visited arcs.

  All the wired-in steps used by the enumeration procedures below are
  view-preserving. */

/* EVEN AND ODD PATHS

  Two arcs {a,b} are /oddly connected/ (resp. /evenly connected/) if
  one can be obtained from the other by an odd (resp. even) number of
  steps, where each step is either {fflip}, {vflip}, or {eflip}
  (where {eflip(a) = fflip(onext(a))}).
  
  A step-func {stp} is /orientation-preserving/ within some submap {N}
  if there is no arc {a} of {N} such that {a} and {stp(a)} are
  oddly-connected. (Note that this is a weaker requirement than `{a}
  and {stp(a)} are evenly connected for every {a}.)

  If {Y} is an orientable connected component of the space {X}, then
  any enumeration that starts from some root {r} on {Y} and uses only
  steps that are orientation-preserving within {Y} will only visit
  arcs that determine the same orientation on {Y} as {r}. In
  particular, if an arc {a} is visited, then {fflip(a)}, {vflip(a)},
  and {eflip(a)} are not; and vice-versa. In this case, the arcs with
  opposite orientation can be be obtained by applying any
  of those flip operations to every visited arc.

  On the other hand, if the the starting root {r} lies on a
  non-orientable connected component {Y} of {X}, then both {a} and one
  of {fflip(a)}, {vflip(a)}, and {eflip(a)} may be visited, for some {a}.
  Ditto if the given root set contains two arcs that are
  oddly-connected to each other.

  Some of the enumeration procedures below use only orientation-preserving
  steps, but some don't.  */

  
  
  
  

*/

typedef void oct_arc_visit_proc_t(oct_arc_t e);
  /* A procedure that can be applied to an arc of an octedge structure. */

uint oct_node_degree(oct_arc_t p);
uint oct_face_degree(oct_arc_t p);

bool_t oct_enum_arcs(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);

bool_t oct_enum_nodes_even(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);
bool_t oct_enum_edges_even(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);
bool_t oct_enum_faces_even(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);

bool_t oct_enum_nodes_any(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);
bool_t oct_enum_edges_any(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);
bool_t oct_enum_faces_any(oct_arcs_t root, oct_visit_t visit, oct_arcs_t *vP);

bool_t oct_enum_ring(oct_arcs_t root, oct_step_t step, oct_visit_t visit, oct_arcs_t *vP);


bool_t oct_enum_orbits(...);
/* 
  A step-func needs not be one-to-one, and may even take a arc in
  {M} to a arc in {M*}, or to a arc that is not connected to {p}
  by any sequence of {flp[i]} and {*} operations. */

/* ---------------------------------------------------------------------- */
/* JUNK? */
/* Enumerate edge quads */
void quad_do_enum (
unsigned next_mark = 1;
void quad_enum(
void quad_do_enum (


void oct_enum_edges(oct_arcs_t *a, oct_arcs_t *et, oct_arc_visit_proc_t *visit);
  /* Visits every arc {e} of the structure that can be reached from the 
    arcs {a->el[0..a->nel-1]} by a chain of {sym} and {onext} operations.
    
    For each new edge record {E} found, sets its numeric identifier to
    a sequential number {id}, starting from 0. Also stores the first
    arc {e} of {E} that was found in the enumeration into
    {et->el[id]}; and, if {visit} is not NULL, calls {visit(e)}. The
    procedure expands and trims {*et} as necessary so that {et->nel}
    will be the number of edges reachable from {a}. */

#endif
