/* See {fn1_zf.h} */
/* Last edited on 2007-02-16 00:11:22 by stolfi */

#include <fn1_zf.h>
#include <fn1_functions.h>

#include <fltgraph.h>
#include <zf.h>
#include <zf_aa.h>
#include <aa.h>
#include <ia.h>
#include <ia_butfly.h>
#include <flt.h>
#include <affirm.h>
#include <jsstring.h>
#include <ps.h>

#include <math.h>
#include <stdio.h>
#include <malloc.h>

#define DEBUG 1

/*** CLOSURE TYPES ***/

typedef struct fn1_zf_problem_t
  { 
    char *title;
    Interval xd;
    Interval yd;
    eval_fp_t *eval_fp;  
    eval_ia_t *eval_ia;
    diff_ia_t *diff_ia;
    eval_aa_t *eval_aa;    
    double epsilon;
    double delta;
    int m;
  } fn1_zf_problem_t;
  
typedef ia_butfly_t eval_butt_t (Interval xv, fn1_zf_problem_t *p);

typedef struct fn1_zf_closure_t
  { FILE *pplot;           /* Progress plot. */
    FILE *splot;           /* Positive/zero/negative plot. */
    eval_butt_t *eval;     /* Function evaluator, returns butterfly enclosure. */
    fn1_zf_problem_t *p; /* Problem description. */
    int nevals;            /* Count of evals performed. */
    int nroots;            /* Number of roots found. */
  } fn1_zf_closure_t;
  

/*** PROTOTYPES FOR INTERNAL PROCEDURES ***/

void fn1_zf_open_ps_files
  ( char *prefix,
    char *tag,
    FILE **pplotp,   /* Zero-finding progress plot. */
    FILE **splotp,   /* Function signs plot. */
    char *extension
  );
  /* Opens Postscript files for the progress plots ({*pplotp}) and
    for the sign (positive/negative/zero) classification of intervals 
    ({*splotp}). The names are "{prefix}-{tag}-{suffix}.{extension}",
    where the suffix is "-p" or "-s", respectively. */

void fn1_zf_plot_preamble
  ( FILE* psfile,
    char *subtitle,      /* Kind of plot (search progress, or function signs). */
    int epsformat,
    char *page,
    char *arithmetic,    /* Kind of arithmetic ("IA" or "AA") */
    fn1_zf_problem_t *p
  );
  /* Writes the preamble of a plot file {psfile}, sets scales, etc. */
  
void fn1_zf_plot_postamble
  ( FILE *psfile,
    int epsformat,
    int nevals,      /* Number of function evaluations (for caption). */
    int nroots,      /* Number of roots found (for caption). */
    fn1_zf_problem_t *p
  );
  /* Writes the postamble of a plot file {psfile}, 
    including the function's plot, axes, frame, captions, etc.. */
  
char *fn1_zf_format_parms
  ( Interval xd,
    Interval yd,
    Float epsilon,
    Float delta
  );
  /* Formats the arguments into a string, for figure captions. */
  
void fn1_zf_report_stats
  ( FILE *pplot, FILE *splot,
    int epsformat, 
    char *arith,    /* "IA" or "AA" */
    int nevals, 
    int nroots
  );
  /* Prints evaluation report, and adds caption lines if not {epsformat} */
  
ia_butfly_t fn1_zf_ia_eval (Interval xv, fn1_zf_problem_t *p);
  /* Computes a rectangular enclosing box for the graph of {F} over the
    interval {xv}. The result has {xmd=xv.lo} and {yxlo=yxmd=yxhi=yv}
    where {yv = p->eval_ia(xv)}. */

ia_butfly_t fn1_zf_ar_eval (Interval xv, fn1_zf_problem_t *p);
  /* Computes a  rectangular enclosing box for the graph of {F} over the
    interval {xv}, from the affine form {p->eval_aa(from_interval(xv))}. */

ia_butfly_t fn1_zf_id_eval (Interval xv, fn1_zf_problem_t *p);
  /* Computes a butterfly-shaped enclosing region for the graph of
    {F}, over the X range {xv}. Uses the interval-slope model with
    central interval {yxmd = p->eval_ia(xm)}, where {xm} is a
    singleton interval at the center {xmd} of {xv}. The slope range is
    obtained by {dv = p->diff_ia(xv)}. */

ia_butfly_t fn1_zf_aa_eval (Interval xv, fn1_zf_problem_t *p);
  /* Computes a parallelogram-shaped enclosure for the graph of {F}, 
    from the affine form {p->eval_aa(from_interval(xv))}. */

void fn1_zf_do_find_and_plot_zeros
  ( FILE *pplot, FILE *splot,
    int epsformat,
    char *page,
    char *arithmetic,
    fn1_zf_problem_t *p,
    eval_butt_t eval
  );
  /* The {eval} function should compute a butterfly containing the 
    graph of the function over the interval {xv}, without any further
    processing. */

/*** IMPLEMENTATIONS ***/

void fn1_zf_find_and_plot_zeros
  ( char *fileprefix,
    int epsformat,
    eval_fp_t eval_fp,
    eval_ia_t eval_ia,
    diff_ia_t diff_ia,
    eval_aa_t eval_aa,
    char *title,
    Interval xd,
    Interval yd,
    Float epsilon,
    Float delta,
    int m
  )
  { FILE *pplot;
    FILE *splot;
    fn1_zf_problem_t p;
    
    p.eval_fp = eval_fp;
    p.eval_ia = eval_ia;
    p.diff_ia = diff_ia;
    p.eval_aa = eval_aa;
    p.title = title;
    p.xd = xd;
    p.yd = yd;
    p.epsilon = epsilon;
    p.delta = delta;
    p.m = m;
    
    if (! epsformat)
      { fn1_zf_open_ps_files(fileprefix, "", &pplot, &splot, ".ps");
        ps_begin_document(pplot, "letter");
        ps_begin_document(splot, "letter");
      }

    /*** Box enclosures (Interval Arith.) ***/
    
    { if (epsformat)
        { fn1_zf_open_ps_files(fileprefix, "ia", &pplot, &splot, ".eps"); }

      fn1_zf_do_find_and_plot_zeros
        ( pplot, splot, epsformat, 
          "ia", "Box enclosures (Interval Arith.)",
          &p, fn1_zf_ia_eval 
        );
    }

    /*** Box enclosures (Affine Arith.) ***/
    { if (epsformat)
        { fn1_zf_open_ps_files(fileprefix, "ar", &pplot, &splot, ".eps"); }

      fn1_zf_do_find_and_plot_zeros
        ( pplot, splot, epsformat,
          "ar", "Box enclosures (Affine Arith.)",
          &p, fn1_zf_ar_eval 
        );
    }

    /*** Interval-slope enclosures (Interval Arith.) ***/
    
    { if (epsformat)
        { fn1_zf_open_ps_files(fileprefix, "id", &pplot, &splot, ".eps"); }

      fn1_zf_do_find_and_plot_zeros
        ( pplot, splot, epsformat, 
          "id", "Interval-slope enclosures (Interval Arith.)",
          &p, fn1_zf_id_eval 
        );
    }

    /*** Slab enclosures (Affine Arith.) ***/
    { if (epsformat)
        { fn1_zf_open_ps_files(fileprefix, "-aa", &pplot, &splot, ".eps"); }

      fn1_zf_do_find_and_plot_zeros
        ( pplot, splot, epsformat,
          "aa", "Slab enclosures (Affine Arith.)",
          &p, fn1_zf_aa_eval 
        );
    }
    
    if (! epsformat)
      { ps_end_document(pplot); fclose(pplot); 
        ps_end_document(splot); fclose(splot);
      }
  }
  
void fn1_zf_open_ps_files
  ( char *prefix,
    char *tag,
    FILE **pplotp,
    FILE **splotp,
    char *extension
  )
  { (*pplotp) = fopen(txtcat(txtcat(prefix, "-p"), extension), "w");
    (*splotp) = fopen(txtcat(txtcat(prefix, "-s"), extension), "w");
  }
  
void fn1_zf_do_find_and_plot_zeros
  ( FILE *pplot,
    FILE *splot,
    int epsformat,
    char *page,
    char *arithmetic,
    fn1_zf_problem_t *p,
    eval_butt_t eval
  )
  { fprintf(stderr, "\nenter fn1_zf_do_find_and_plot_zeros...\n");

    int nevals = 0;
    int nroots = 0;
    
    auto void fn1_zf_eval(Interval *x, Interval *y, ia_butfly_t *bt);
      /* Called by {zf_enum_zeros} to bound the graph of {F} and check for
        roots. This procedure calls {closure->eval(xv, closure->p)} and
        plots the resulting butterfly. The {data} argument should be
        actually an {fn1_zf_closure_t}. */

    auto bool_t fn1_zf_report(Interval *x, Interval *y, zf_kind_t kind);
      /* Called by {zf_enum_zeros} to report a new interval.
        This procedure plots the given root and prints it to {stderr}. */
    
    fn1_zf_plot_preamble
      (pplot, "Search progress",  epsformat, page, arithmetic, p);
    fn1_zf_plot_preamble
      (splot, "Function signs",   epsformat, page, arithmetic, p);

    ps_set_pen(pplot, 0.0, 0.5, 0.0, 0.25, 0.0, 0.0);
    ps_set_pen(splot, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0);
    zf_enum_zeros
      ( fn1_zf_eval, 
        fn1_zf_report,
        p->xd,
        p->epsilon,
        p->delta
      );
    
    fn1_zf_plot_postamble(pplot, epsformat, nevals, nroots, p);
    fn1_zf_plot_postamble(splot, epsformat, nevals, nroots, p);
    fprintf(stderr, "exit fn1_zf_do_find_and_plot_zeros.\n");
    return;
    
    /* Internal procedures: */

    void fn1_zf_eval (Interval *x, Interval *y, ia_butfly_t *bt)
      { *bt = eval(*x, p);
        Interval ybt0 = ia_join(bt->tp[0].yxlo, bt->tp[0].yxhi);
        Interval ybt1 = ia_join(bt->tp[1].yxlo, bt->tp[1].yxhi);
        *y = ia_join(ybt0, ybt1);
        nevals++;
        ia_butfly_draw(pplot, &(p->yd), bt);
      }

    bool_t fn1_zf_report(Interval *x, Interval *y, zf_kind_t kind)
      { nroots++;
        fprintf(stderr, "  ** apparent root in "); 
        ia_print(stderr, *x); 
        fprintf(stderr, "\n");

        { /* Plot result: */
          float cr, cg, cb; /* Polygon color */
          switch (kind)
            { case zf_kind_undefined:
                cr = 0.75; cg = 0.75; cb = 0.75; break;
              case zf_kind_positive:
                cr = 1.00; cg = 0.25; cb = 0.25; break;
              case zf_kind_negative:
                cr = 0.50; cg = 0.60; cb = 1.00; break;
              case zf_kind_root:
                cr = 0.00; cg = 0.85; cb = 0.00; break;
              case zf_kind_mixed:
                cr = 0.00; cg = 0.00; cb = 0.00; break;
            }

          ia_trapez_t tr = (ia_trapez_t){ *x, *y, *y };
          ia_trapez_fill(splot, &(p->yd), &tr, cr,cg,cb);
        }
        return 0;
      }
  }
  
void fn1_zf_plot_preamble
  ( FILE* psfile,
    char *subtitle,
    int epsformat,
    char *page,
    char *arithmetic,
    fn1_zf_problem_t *p
  )
  { char *parmstr = fn1_zf_format_parms(p->xd, p->yd, p->epsilon, p->delta);
    Float wx = p->xd.hi - p->xd.lo;
    Float wy = p->yd.hi - p->yd.lo;
    Float wm = FMAX(wx, wy);
    double scale = 6.0 * 72.0 / wm;
    double xc = 4.25 * 72.0;
    double hmin = xc - scale * wx / 2.0;
    double hmax = xc + scale * wx / 2.0;
    double yc = 6.50 *72.0;
    double vmin = yc - scale * wy / 2.0;
    double vmax = yc + scale * wy / 2.0;

    if (psfile == NULL)
      { fatalerror ("fn1_zf_do_find_and_plot_zeros: null PostScript file"); }

    if (epsformat)
      { ps_begin_figure(psfile, hmin, hmax, vmin, vmax);
        ps_set_window(psfile,
          p->xd.lo, p->xd.hi, p->yd.lo, p->yd.hi,
          hmin,  hmax,  vmin,  vmax,
          1, 1
        );
        ps_set_pen(psfile, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0);
      }
    else
      { ps_begin_page(psfile, page);
	ps_set_window(psfile,
          p->xd.lo, p->xd.hi, p->yd.lo, p->yd.hi,
          hmin,  hmax,  vmin,  vmax,
          1, 1
        );
        ps_set_pen(psfile, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0);
        ps_add_caption(psfile, p->title);
        ps_add_caption(psfile, subtitle);
        ps_add_caption(psfile, txtcat("Arithmetic: ", arithmetic));
        ps_add_caption(psfile, "");
        ps_add_caption(psfile, parmstr);
      }

    ps_comment(psfile, arithmetic);
  }

void fn1_zf_plot_postamble
  ( FILE *psfile,
    int epsformat,
    int nevals,
    int nroots,
    fn1_zf_problem_t *p
  )
  { ps_set_pen(psfile, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0);

    { char sevals[80];
      sprintf(sevals, 
        "%d function evaluations, %d roots found", 
        nevals, nroots
      );
      fprintf(stderr, "%s\n", sevals);
      if (! epsformat) { ps_add_caption(psfile, sevals); }
    }
    
    ps_set_pen(psfile, 0.0, 0.0, 0.0, 0.10, 0.0, 0.0);
    fltgraph_draw_axes(psfile, p->xd, p->yd);
    
    ps_set_pen(psfile, 0.5, 0.0, 0.0, 0.35, 0.0, 0.0);
    fltgraph_plot(psfile, p->eval_fp, p->xd, p->yd, p->m);
    
    ps_set_pen(psfile, 0.0, 0.0, 0.0, 0.25, 0.0, 0.0);
    ps_draw_frame(psfile);

    if (epsformat)
      { ps_end_figure(psfile); }
    else
      { ps_end_page(psfile); }
  }

ia_butfly_t fn1_zf_ia_eval (Interval xv, fn1_zf_problem_t *p)
  { Interval fv = p->eval_ia(xv);
#   if DEBUG
      fprintf(stderr, "  f =  "); ia_print(stderr, fv); fprintf(stderr, "\n");
#   endif
    return ia_butfly_from_box(&xv, &fv);
  }

ia_butfly_t fn1_zf_id_eval (Interval xv, fn1_zf_problem_t *p)
  { /* Compute Y range at central point of {xv}: */
    Float xmd = ia_mid(xv);
    Interval xm = (Interval){xmd, xmd};
    Interval yxmd = p->eval_ia(xm);
    /* Compute slope range over whole range {xv}: */
    Interval dfv = p->diff_ia(xv);
#   if DEBUG
      fprintf(stderr, "  f =  "); ia_print(stderr, yxmd); fprintf(stderr, "\n");
      fprintf(stderr, "  df = "); ia_print(stderr, dfv); fprintf(stderr, "\n");
#   endif
    return ia_butfly_from_ia_diff(&xv, xmd, &yxmd, &dfv);
  }

ia_butfly_t fn1_zf_aa_eval (Interval xv, fn1_zf_problem_t *p)
  { MemP frame = aa_top();
    AAP xf = aa_from_interval(xv);
    AAP yf = p->eval_aa(xf);
#   if DEBUG
      fprintf(stderr, "  f =  "); aa_print(stderr, yf); fprintf(stderr, "\n");
#   endif

    ia_trapez_t tr = zf_trapezoid_from_aa(&xv, xf, yf);
    aa_flush(frame);
    return ia_butfly_from_trapez(&tr);
  }

ia_butfly_t fn1_zf_ar_eval (Interval xv, fn1_zf_problem_t *p)
  { MemP frame = aa_top();
    AAP xf = aa_from_interval(xv);
    AAP yf = p->eval_aa(xf);
#   if DEBUG
      fprintf(stderr, "  f =  "); aa_print(stderr, yf); fprintf(stderr, "\n");
#   endif
    Interval yv = aa_range(yf);
    ia_butfly_t bt = ia_butfly_from_box(&xv, &yv);
    aa_flush(frame);
    return bt;
  }

char *fn1_zf_format_parms
  ( Interval xd,
    Interval yd,
    Float epsilon,
    Float delta
  )
  { char *s = (char *) malloc(100);
    sprintf(s, 
      "x in [%f _ %f]  y in [%f _ %f]\nepsilon = %.4e delta = %.4e",
      xd.lo, xd.hi, yd.lo, yd.hi, epsilon, delta
    );
    return s;
  }

